/*
 * trnames2.c - new-fashioned filename translation routines
 * 
 * Michal Ludvig <michal@logix.cz> (c) 2001
 * 
 * tr2_getfarray()    will return all valid entries from 'dirname'
 *                    which match 'filename'-s base in an array,
 *                    sorted by prefference. Array is variable-
 *                    -length and is ended by a NULL pointer.
 * tr2_freefarray()   walk through farray returned by tr2_getfarray() 
 *                    and free() all it's items. Finally free() the 
 *                    array itself too.
 * tr2_xlate4create() does a filename translation for file/directory 
 *                    creations requests. Uses file.ext$$CREATE=...$$
 * tr2_reducename()   called from readdir to show only matching files.
 */

#define _BSD_SOURCE

#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <stdlib.h>
#include <dirent.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <rpc/rpc.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "trnames2.h"

#define	BUFSIZE	1000

#define	_CNFS_CR_NUM_SCOPES		2
#define	_CNFS_CR_SCOPE_SPECIFIC	0
#define	_CNFS_CR_SCOPE_GENERAL	1

#define	FNYES	0
#define	FNNO	1

#define	_LOG_SYSLOG
// #define	_LOG_STDERR

typedef char*	filename;

static void logmsg(const char *format, ...)
{
	va_list	args;
	char		buf[BUFSIZE+1];
#ifdef STANDALONE
	static int	logopened=0;
#endif
	
	va_start(args, format);
	vsnprintf(buf, BUFSIZE, format, args);

#ifdef _LOG_SYSLOG
#ifdef STANDALONE
	if(!logopened) { openlog("test1", LOG_PID, LOG_DAEMON); logopened=1; }
#endif /* STANDALONE */

	syslog(LOG_DAEMON|LOG_INFO, "%s", buf);
#endif

#ifdef _LOG_STDERR
	fprintf(stderr, "%s\n", buf);
#endif
}

static int fnmatch(const char *fname, struct filelist *retflist, struct svc_req *rqstp)
{
	struct filelist	flist, *pflist;
	struct in_addr		inaddr;
	struct authunix_parms *unix_cred;
	char	*p1, *p2, *name, tag[NAME_MAX+1], val[NAME_MAX+1],
			ipaddr[NAME_MAX], netmask[NAME_MAX];
	int	fnlen, len, currtag, valid;
	
	
	pflist=&flist;
	memset(pflist, 0, sizeof(struct filelist));
	
	name=strdup(fname);
	p1=strstr(name, "$$");
	
	if(! p1)
	{
		snprintf(pflist->fnbase, NAME_MAX, "%s", fname);
		pflist->tags|=CNFS_TAG_PURE;
		if(retflist) memcpy(retflist, pflist, sizeof(struct filelist));
		return FNYES;
	}
		
	*p1=0;
	p1+=2;
		
	fnlen=strlen(p1);
	if(p1[fnlen-1]!='$' || p1[fnlen-2]!='$')
	{
		/* doesn't end with $$ - print as it is */
		snprintf(pflist->fnbase, NAME_MAX, "%s", fname);
		pflist->tags|=CNFS_TAG_PURE;
		if(retflist) memcpy(retflist, pflist, sizeof(struct filelist));
		return FNYES;
	}
	
	p1[fnlen-2]=0;
	fnlen-=2;
	snprintf(pflist->fnbase, NAME_MAX, "%s", name);
	snprintf(pflist->fntag, NAME_MAX, "%s", p1);
		
	p2=p1;	/* let following condition pass at least once */
		
	while(*(p1=p2))
	{
		p2=strchr(p1, ',');
		if(p2) *p2++=0;
			else p2=&p1[strlen(p1)];
#ifdef CONFIG_TAG_CLIENT
		if(strcmp(p1, "CLIENT")==0 || strcmp(p1, "CLIENTS")==0)
		{
			pflist->tags|=CNFS_TAG_CLIENT;
			continue;
		}
#endif
		if(sscanf(p1, "%[A-Z]=%s", tag, val)==2)
		{
			currtag=0;

			if(strcasecmp(tag, "IP")==0)		currtag=CNFS_TAG_IP;
			if(strcasecmp(tag, "UID")==0)		currtag=CNFS_TAG_UID;
			if(strcasecmp(tag, "GID")==0)		currtag=CNFS_TAG_GID;
			if(strcasecmp(tag, "GROUP")==0)	currtag=CNFS_TAG_GROUP;
			if(strcasecmp(tag, "HOST")==0)	currtag=CNFS_TAG_HOST;
			if(strcasecmp(tag, "CREATE")==0)	currtag=CNFS_TAG_CREATE;

			len=strlen(val)+1;

			if(currtag==0 || len==1)
			{
//				logmsg("Invalid tag '%s=%s' at %s/%s", tag, val, dirname, fname);
				logmsg("Invalid tag '%s=%s' at %s", tag, val, fname);
				continue;
			}

			switch(currtag)
			{
#ifdef CONFIG_TAG_IP
				case CNFS_TAG_IP:
					*ipaddr=0;
					*netmask=0;
					if(strchr(val, '_'))
					{
						if(sscanf(val, "%[0-9.]_%[0-9.]", ipaddr, netmask)!=2)
							continue;
					}
					else
					{
						if(sscanf(val, "%[0-9.]", ipaddr)!=1)
							continue;
					}
					
					if(inet_aton(ipaddr, &inaddr)==0)
						continue;
					
					pflist->ipaddr=ntohl(inaddr.s_addr);
					pflist->netmask=-1;
					
					if(*netmask)
					{
						if(strchr(netmask, '.'))
						{
							inet_aton(netmask, &inaddr);
							pflist->netmask=ntohl(inaddr.s_addr);
						}
						else
						{
							sscanf(netmask, "%d", &len);
							if(len>=0 && len<=32)
								pflist->netmask=0xFFFFFFFF00000000L>>len;
						}
						pflist->tags|=CNFS_TAG_NET;
					}
					else
						pflist->tags|=CNFS_TAG_IP;
					break;
#endif

#ifdef CONFIG_TAG_UID
				case CNFS_TAG_UID:
					if(sscanf(val, "%d", &pflist->uid)==1)
					{
						logmsg("%s: UID: %d", pflist->fntag, pflist->uid);
						pflist->tags|=CNFS_TAG_UID;
					}
					break;
#endif

#ifdef CONFIG_TAG_GID
				case CNFS_TAG_GID:
					if(sscanf(val, "%d", &pflist->gid)==1)
					{
						logmsg("%s: GID: %d", pflist->fntag, pflist->gid);
						pflist->tags|=CNFS_TAG_GID;
					}
					break;
#endif

#ifdef CONFIG_TAG_GROUP
				case CNFS_TAG_GROUP:
					pflist->group=calloc(1, len);
					snprintf(pflist->group, len, "%s", val);
					logmsg("%s: GROUP: %s", pflist->fntag, pflist->group);
					pflist->tags|=CNFS_TAG_GROUP;
					break;
#endif

#ifdef CONFIG_TAG_HOST
				case CNFS_TAG_HOST:
					pflist->host=calloc(1, len);
					snprintf(pflist->host, len, "%s", val);
					pflist->tags|=CNFS_TAG_HOST;
					break;
#endif

				default:
//					logmsg("%s: Non-fatal internal error: Unknown but parsed TAG: 0x%04X", pflist->fnbase, currtag);
					break;
			}
		}
	}
	
	/* create convenient pointer to RPC credidentials */
	unix_cred = (struct authunix_parms *) rqstp->rq_clntcred;

	valid=1;
		
#ifdef CONFIG_TAG_UIP
	if(pflist->tags & CNFS_TAG_UID)
	{
		if(pflist->uid != unix_cred->aup_uid)
			valid=0;
		else
			pflist->match++;
	}
#endif

#ifdef CONFIG_TAG_GID
	if(pflist->tags & CNFS_TAG_GID)
	{
		if(pflist->gid != unix_cred->aup_gid)
			valid=0;
		else
			pflist->match++;
	}
#endif

#ifdef CONFIG_TAG_IP
	if(pflist->tags & CNFS_TAG_IP || pflist->tags & CNFS_TAG_NET)
	{
		if((ntohl(svc_getcaller(rqstp->rq_xprt)->sin_addr.s_addr) & pflist->netmask) == (pflist->ipaddr & pflist->netmask))
			pflist->match++;
		else
			valid=0;
	}
#endif

#ifdef CONFIG_TAG_HOST
	if(pflist->tags & CNFS_TAG_HOST)
	{
		int len, wildcard=0;
		len=strlen(pflist->host);
		if(pflist->host[len-1]=='*')
		{
			wildcard=1;
			pflist->host[len-1]=0;
			len--;
		}

		if(wildcard)
		{
			pflist->tags &= ~CNFS_TAG_HOST;
			pflist->tags |= CNFS_TAG_HOSTW;
		}
		
		if((wildcard && strncasecmp(pflist->host, unix_cred->aup_machname, len)==0) ||
			strcasecmp(pflist->host, unix_cred->aup_machname)==0)
			pflist->match++;
		else
			valid=0;
	}
#endif

#ifdef CONFIG_TAG_GROUP
	#error "CONFIG_TAG_GROUP Not yet supported"
#endif

#ifdef CONFIG_TAG_PURE
	if(pflist->tags & CNFS_TAG_PURE)
		pflist->match++;
#endif

#ifdef CONFIG_TAG_CLIENT
	if(pflist->tags & CNFS_TAG_CLIENT)
		pflist->match++;
#endif

	if(valid==0)
		pflist->match=0;

	if(retflist) memcpy(retflist, pflist, sizeof(struct filelist));
	return pflist->match>0?FNYES:FNNO;
}

int tr2_xlate4create(const char *dirname, char *filename, struct svc_req *rqstp)
{
	DIR		*dir;
	char		*p1, *p2, newname[NAME_MAX], *c_tag="$$CREATE=", pattern[NAME_MAX];
	int		c_tag_len, fnlen, patt_len, currscope;
	_u32		tags[_CNFS_CR_NUM_SCOPES]={0};
	struct dirent		*dent;
	struct authunix_parms *unix_cred;
	
	c_tag_len=strlen(c_tag);
	patt_len=snprintf(pattern, NAME_MAX, "%s%s", filename, c_tag);
	
	/* create convenient pointer to RPC credidentials */
	unix_cred = (struct authunix_parms *) rqstp->rq_clntcred;
	
	/* Open directory and return just a requested filename in case of failure. */
	if(! (dir=opendir(dirname)))
	{
		syslog(LOG_WARNING, "opendir(%s): %s", dirname, strerror(errno));
		return -1;
	}
	
	/* Now walk through the directory and put entries 
	 * beginning with 'filename' to the list. 
	 * Also parse all tags and store values into filelist chain */
	while((dent=readdir(dir)))
	{
		p1=dent->d_name;
		p2=strstr(p1, c_tag);
		
		if(!p2)
			continue;
		
		if(p2==p1)
			currscope=_CNFS_CR_SCOPE_GENERAL;
		else if(strncmp(pattern, p1, patt_len)==0)
			currscope=_CNFS_CR_SCOPE_SPECIFIC;
		else
			continue;
		
		p1=p2;
		p1+=c_tag_len;
		fnlen=strlen(p1);
		if(p1[fnlen-1]!='$' || p1[fnlen-2]!='$')
			continue;
		
		p1[fnlen-2]=0;
		fnlen-=2;
		while(*p1)
		{
			p2=strchr(p2, ',');
			if(p2) *p2++=0;
			
			if(strcasecmp(p1, "IP")==0)		tags[currscope]|=CNFS_TAG_IP;
			if(strcasecmp(p1, "UID")==0)		tags[currscope]|=CNFS_TAG_UID;
			if(strcasecmp(p1, "GID")==0)		tags[currscope]|=CNFS_TAG_GID;
			if(strcasecmp(p1, "GROUP")==0)	tags[currscope]|=CNFS_TAG_GROUP;
			if(strcasecmp(p1, "HOST")==0)		tags[currscope]|=CNFS_TAG_HOST;
			if(strcasecmp(p1, "CLIENT")==0)	tags[currscope]|=CNFS_TAG_CLIENT;
			
			if(p2) p1=p2;
			else break;
		}
	}
	
	closedir(dir);
	
	newname[0]=0;
	
	for(currscope=0; currscope<_CNFS_CR_NUM_SCOPES; currscope++)
	{
		char	newtag[NAME_MAX]={0};
		
		if(! tags[currscope])
			continue;
	
#ifdef CONFIG_TAG_IP
		if(tags[currscope] & CNFS_TAG_IP)
			snprintf(newtag, NAME_MAX, "%s,IP=%s", newtag, inet_ntoa(svc_getcaller(rqstp->rq_xprt)->sin_addr));
#endif
		
#ifdef CONFIG_TAG_UID
		if(tags[currscope] & CNFS_TAG_UID)
			snprintf(newtag, NAME_MAX, "%s,UID=%d", newtag, unix_cred->aup_uid);
#endif
		
#ifdef CONFIG_TAG_GID
		if(tags[currscope] & CNFS_TAG_GID)
			snprintf(newtag, NAME_MAX, "%s,GID=%d", newtag, unix_cred->aup_gid);
#endif
		
#ifdef CONFIG_TAG_GROUP
	#warning	"tr2_xlate4create(): CONFIG_TAG_GROUP not yet implemented"
#endif
		
#ifdef CONFIG_TAG_HOST
		if(tags[currscope] & CNFS_TAG_HOST)
			snprintf(newtag, NAME_MAX, "%s,HOST=%s", newtag, unix_cred->aup_machname);
#endif
		
#ifdef CONFIG_TAG_CLIENT
		if(tags[currscope] & CNFS_TAG_CLIENT)
			snprintf(newtag, NAME_MAX, "%s,CLIENT", newtag);
#endif
		
		if(newtag[0]==',')
		{
			snprintf(newname, NAME_MAX, "%s$$%s$$", filename, &newtag[1]);
			break;
		}
	}
	
	if(newname[0])
		snprintf(filename, NAME_MAX, "%s", newname);
	
	return 0;
}

struct filelist **tr2_getfarray(const char *dirname, const char *filename, struct svc_req *rqstp)
{
	DIR	*dir;
	char	*p1, fname[NAME_MAX+1];
	int	i, j, fnlen, maxmatch, havemore, currid, 
			entries, entry1, entry2, maxtag;
	struct dirent		*dent;
	struct filelist	*flist, *pflist, *prevflist, **arr1, **arr2;
	
	if(! rqstp)
	{
		logmsg("Internal error: tr2_getflist() called with rqstp==NULL");
		return NULL;
	}
	
	snprintf(fname, NAME_MAX, "%s", filename);
	
	flist=calloc(1, sizeof(struct filelist));
	pflist=flist;
	if(! pflist) { logmsg("calloc(): %s", strerror(errno)); return NULL; }
	
	/* Open directory and return just a requested filename in case of failure. */
	if(! (dir=opendir(dirname)))
	{
		syslog(LOG_WARNING, "opendir(%s): %s", dirname, strerror(errno));
		arr2=calloc(2, sizeof(struct filelist *));
		arr2[0]=calloc(1, sizeof(struct filelist));
		snprintf(arr2[0]->fnbase, NAME_MAX, "%s", filename);
		return arr2;
	}
	
	/* Now walk through the directory and put entries 
	 * beginning with 'filename' to the list. 
	 * Also parse all tags and store values into filelist chain */
	while((dent=readdir(dir)))
	{
		p1=dent->d_name;
		if(pflist->fnbase[0]!=0)
		{
			pflist->next=calloc(1, sizeof(struct filelist));
			pflist=pflist->next;
			if(! pflist) { logmsg("calloc(): %s", strerror(errno)); closedir(dir); return NULL; }
		}
		
		fnlen=strlen(fname);
		if(strncmp(fname, p1, fnlen)!=0)
			continue; 
		if(p1[fnlen]!=0 && (p1[fnlen]!='$' || p1[fnlen+1]!='$'))
			continue;
		fnmatch(p1, pflist, rqstp);
	}
	
	closedir(dir);
	
	/* Now, when we have read all the directory, 
	 * check each item for validity */
	pflist=flist;
	maxmatch=0;
	
	while(pflist) 
	{
		if(pflist->match > maxmatch)
			maxmatch=pflist->match;
		
		/* get next item */
		pflist=pflist->next;
	}
	
	/* count valid entries */
	pflist=flist;
	entries=0;
	while(pflist)
	{
		entries++;
		pflist=pflist->next;
	}
	
	arr1=(struct filelist **)calloc(entries+1, sizeof(struct filelist *));
	arr2=(struct filelist **)calloc(entries+1, sizeof(struct filelist *));
	
	/* move all entries to an array */
	pflist=flist;
	entry1=0;
	while(pflist)
	{
		arr1[entry1]=malloc(sizeof(struct filelist));
		bcopy(pflist, arr1[entry1], sizeof(struct filelist));
		arr1[entry1]->next=NULL;
		entry1++;
		prevflist=pflist;
		pflist=pflist->next;
		free(prevflist);
	}
	
	entry2=0;
	
	/* decrement 'i' only when we don't have more entries with match==i */
	for(i=maxmatch; i>0; i=havemore?i:i-1)	
	{
		havemore=0;
		maxtag=0;
		currid=-1;
		
		for(j=0; j<entries; j++)
		{
			if(! arr1[j])
				continue;
			
			if(arr1[j]->match == 0)
			{
				free(arr1[j]);
				arr1[j]=NULL;
				continue;
			}
			
			if(arr1[j]->match == i)
			{
				if(arr1[j]->tags > maxtag)
				{
					maxtag=arr1[j]->tags;
					currid=j;
				}
				
				havemore++;
			}
		}
		
		if(currid>=0)
		{
			arr2[entry2++]=arr1[currid];
			arr1[currid]=NULL;
		}
	}
	
	/* free all remaining items from arr1[] and then arr1[] too */
	for(i=0; i<entries; i++)
		if(arr1[i])
			free(arr1[i]);
	free(arr1);
	
	return arr2;
}

void tr2_freefarray(struct filelist **farray)
{
	int	i;
	
	for(i=0; farray[i]; i++)
		if(farray[i])
			free(farray[i]);
	free(farray);
	farray=NULL;
}

char *tr2_reducename(char *filename, struct svc_req *rqstp)
{
	static struct filelist	flist;
	int	len;
	char	*name;
	
	if(! filename) return NULL;
	
	name = strdup(filename);
	len = strlen(name);
	
	if(name[len-2] != '$' || name[len-1] != '$') return name; /* must end with $$ */
	if(len < 5) return name;	/* must be at least <char>$$$$ */
	
	memset(&flist, 0, sizeof(struct filelist));
	
	if(fnmatch(name, &flist, rqstp)==FNNO)
		return NULL;
	else
		return flist.fnbase;
}
